/*
 * main.c -- the core of the ADL7200 driver
 *
 * Copyright (C) 1998   rubini@linux.it (Alessandro Rubini)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This driver only works with Linux 2.0 and 2.1.43, by now */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/bios32.h> /* pcibios_* */
#include <linux/pci.h>
#include <linux/delay.h>

#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/io.h>

#include "adl7200.h"        /* local definitions */
#include "allocator.h"

/*
 * I don't use static symbols here, because modularization code hides
 * public symbols
 */

int adl_major =   ADL_MAJOR;
int adl_alltimeout = 5; /* timeout for the allocator, if used */

Adl_Dev adl_devices[ADL_MAX_DEV];

/*
 * Different minors behave differently, so let's use multiple fops
 */

struct file_operations *adl_fops[]={
    &adl_raw_fops,      /* type 0: raw I/O */
    &adl_dma_fops,      /* type 1: DMA-driven */
    &adl_trg_fops,      /* type 2: REQ/ACQ triggered */
    &adl_sft_fops,      /* type 3: soft timing */
    &adl_asc_fops,      /* type 4: ascii control node */
};

/*-------------------------------------------------------- timer function */
/*
 * The timer function serves two tasks: it is the engine for adlS devices
 * and awakes processes using the adlD and adlT nodes if DMA is proceeding
 */
void adl_tfun(unsigned long arg)
{
    Adl_Dev *dev = (Adl_Dev *)arg;
    u32 tmp;

    /* FIXME: the soft node is not active with the timer */

    tmp = ADL_IN32(dev,AMCC_REG_MRTC);
    if (tmp && tmp < dev->wpending) {
        PDEBUGG("timer: wake(writer): pending= %i\n",tmp);
        dev->wpending = tmp;
        wake_up_interruptible(&dev->outq);
    }
    tmp = ADL_IN32(dev,AMCC_REG_MWTC);
    if (tmp && tmp < dev->rpending) {
        PDEBUGG("timer: wake(reader): pending= %i\n",tmp);
        dev->rpending = tmp;
        wake_up_interruptible(&dev->inq);
    }
    /* not check over and underruns */
    tmp = ADL_IN32_L(dev,ADL_REG_DIO_CTRL);
    if (tmp & ADL_BIT_OVER)  {
        dev->fifo_overr_count++;
        PDEBUG("timer: overrun (pending %lx)\n",dev->rpending);
    }
    if (tmp & ADL_BIT_UNDE)  {
        PDEBUG("timer: underrun (pending %lx)\n",dev->wpending);
        dev->fifo_under_count++;
    }
    tmp |= ADL_BIT_OVER | ADL_BIT_UNDE;
    ADL_OUT32_L(dev,tmp,ADL_REG_DIO_CTRL);

    if (dev->wpending || dev->rpending) { /* check again */
        del_timer(&dev->timer);
        dev->timer.expires = jiffies+HZ/(dev->hertz);
        add_timer(&dev->timer);
    }
    
    return;
}

/*-------------------------------------------------------- init the device */
/*
 * Device initialization. Before calling this function, at power on,
 * the registers are all set to 0. The received pointer holds
 * valid bus and device fields
 */
int adl_init(Adl_Dev *dev)
{
    u16 cmd;
    u32 tmp;

    /* paranoia check: two regions must be I/O, the third must unexsist */
    pcibios_read_config_dword(dev->bus, dev->device, PCI_BASE_ADDRESS_0,
			      &dev->io_amcc);
    pcibios_read_config_dword(dev->bus, dev->device, PCI_BASE_ADDRESS_1,
			      &dev->io);
    pcibios_read_config_dword(dev->bus, dev->device, PCI_BASE_ADDRESS_2,
			      &tmp);

    if ( ((dev->io & PCI_BASE_ADDRESS_SPACE) != PCI_BASE_ADDRESS_SPACE_IO)
	|| ((dev->io_amcc & PCI_BASE_ADDRESS_SPACE)
	    != PCI_BASE_ADDRESS_SPACE_IO)
	|| (tmp != (u32)0) ) {
	printk(KERN_WARNING ADL_MSG "incorrect I/O regions\n");
	PDEBUG("(%08x, %08x, %08x)\n",(int)dev->io,(int)dev->io_amcc,(int)tmp);
	dev->io = dev->io_amcc = 0UL;
	return -ENODEV;
    }
    dev->io      &= PCI_BASE_ADDRESS_IO_MASK;
    dev->io_amcc &= PCI_BASE_ADDRESS_IO_MASK;

    /* activate it */
    pcibios_read_config_word(dev->bus, dev->device, PCI_COMMAND, &cmd);
    pcibios_write_config_word(dev->bus, dev->device, PCI_COMMAND,
                               cmd | PCI_COMMAND_IO);

    pcibios_read_config_byte(dev->bus, dev->device, PCI_INTERRUPT_LINE,
			     &dev->irq);

    init_timer(&dev->timer);
    dev->timer.function = adl_tfun;
    dev->timer.data = (unsigned long)dev;

    /*
     * Should make additional checks? The AMCC chip is quite general...
     * Well, it *looks* like the device ID is unique for the ADL7200
     */

    return 0; /* success */
}

/*---------------------------------------------------- seting up DMA */
void adl_stop_dma(Adl_Dev *dev)
{
    u32 tmp;
    PDEBUGG("stopping DMA\n");

    /* stop both in and out DMA's (but don't touch dev->flags) */
    tmp = ADL_IN32(dev,ADL_REG_DIO_CTRL);

    /* DI & DO */
    tmp &= ~ADL_MASK_DISABLE ;
    ADL_OUT32(dev, tmp, ADL_REG_DIO_CTRL);

    tmp = ADL_IN32(dev, AMCC_REG_INTCSR);
    tmp &= ~(AMCC_BIT_INTCSR_DOWRINT | AMCC_BIT_INTCSR_DORDINT);
    tmp &= 0xFF00FFFF; /* ? */
    ADL_OUT32(dev, tmp, AMCC_REG_INTCSR);

    tmp = ADL_IN32(dev, AMCC_REG_MCSR);
    tmp &= ~(AMCC_BIT_MSCR_DOWRREQ | AMCC_BIT_MSCR_DORDREQ);
    tmp &= ~(AMCC_BIT_MCSR_RSTWRFIFO | AMCC_BIT_MCSR_RSTRDFIFO);
    ADL_OUT32(dev, tmp, AMCC_REG_INTCSR);

#if 0
    tmp = ADL_IN32(dev, AMCC_REG_MRTC);
    PDEBUG("pending write xfers (bytes): %i\n",(int)tmp);
    tmp = ADL_IN32(dev, AMCC_REG_MWTC);
    PDEBUG("pending read xfers (bytes): %i\n",(int)tmp);
#endif
    return;
}

/*
 * This function is in charge of setting up DMA by talking with
 * the 5933. All information is received from the Adl_Dev structure
 */
int adl_run_dma(Adl_Dev *dev, u32 bitstoset, u32 bitstoclear)
{
    u32 tmp, flags_intcsr=0, flags_mcsr=0;

    adl_stop_dma(dev);
    PDEBUGG("setting up DMA (flags 0x%08lx)\n",dev->flags);
    if (dev->flags & ADL_FLAG_READING) { /* i.e., write to PCI */
	ADL_OUT32_A(dev, dev->dmabuffer, AMCC_REG_MWAR);
	ADL_OUT32_A(dev, dev->dma_requested_size, AMCC_REG_MWTC);
        dev->rpending = dev->dma_requested_size;
	/* the bits use RD and WR the way applications see them */
	flags_intcsr = AMCC_BIT_INTCSR_DORDINT;
	flags_mcsr = AMCC_BIT_MCSR_RSTRDFIFO | AMCC_BIT_MSCR_DORDREQ
	    | AMCC_BIT_MSCR_RDPRIO;
    }
    if (dev->flags & ADL_FLAG_WRITING) { /* i.e., read from PCI */
	ADL_OUT32_A(dev, dev->dmabuffer, AMCC_REG_MRAR);
	ADL_OUT32_A(dev, dev->dma_requested_size, AMCC_REG_MRTC);
        dev->wpending = dev->dma_requested_size;
	flags_intcsr = AMCC_BIT_INTCSR_DOWRINT;
	flags_mcsr = AMCC_BIT_MCSR_RSTWRFIFO | AMCC_BIT_MSCR_DOWRREQ
	    | AMCC_BIT_MSCR_WRPRIO;
    }

    /* now, tell the add-on logic what must be used */
    tmp = ADL_IN32_L(dev,ADL_REG_DIO_CTRL);
    ADL_OUT32_L(dev, (tmp | bitstoset) & ~bitstoclear, ADL_REG_DIO_CTRL);
    PDEBUGG("ADD-ON instructed (0x%08x)\n",
	   (tmp | bitstoset) & ~bitstoclear);

    /* and enable bus mastering */
    tmp = ADL_IN32_A(dev,AMCC_REG_MCSR);
    dev->master_flags = flags_mcsr;
    ADL_OUT32_A(dev, tmp | flags_mcsr, AMCC_REG_MCSR);
    PDEBUGG("MCSR instructed (0x%08x)\n",tmp|flags_mcsr);

    /* finally, enable interrupts */
    tmp = ADL_IN32_A(dev,AMCC_REG_INTCSR);
    dev->irq_flags = flags_intcsr;
    ADL_OUT32_A(dev, tmp | flags_intcsr, AMCC_REG_INTCSR);
    PDEBUGG("enabled irq\n");

    return 0;
}


/*---------------------------------------------------- interrupt handler */
static void adl_interrupt(int irq, void *devptr, struct pt_regs *regs)
{
    Adl_Dev *dev=devptr;
    u32 int_status, master_status;
    /*
     * First of all, we should check if the interrupt is valid.
     * A shared handler is invoked on any interrupt of any device,
     * so we must validate it.
     */
    
    int_status =    ADL_IN32_A(dev,AMCC_REG_INTCSR);
    master_status = ADL_IN32_A(dev,AMCC_REG_MCSR);
    PDEBUGG("isr: intstatus %08lx, mstatus %08lx\n",
	   (unsigned long)int_status, (unsigned long)master_status);
    if ( (int_status & AMCC_BIT_INTCSR_ISMYINT) == 0)
        return; /* not mine */

    /* check what it is */
    if (int_status & AMCC_BIT_INTCSR_ISWRINT) { /* write is done */
	PDEBUGG("isr: write terminated\n");
	dev->dma_count++;
        dev->wpending=0;
	dev->flags &= ~ADL_FLAG_WRITING;
	/* tell the device it's done and acknowledge IRQ */
	ADL_OUT32_A(dev, master_status & ~AMCC_BIT_MSCR_DOWRREQ,
		    AMCC_REG_MCSR);
	ADL_OUT32_A(dev, int_status | AMCC_BIT_INTCSR_ISMYINT
		    | AMCC_BIT_INTCSR_ISWRINT, AMCC_REG_INTCSR);
	wake_up_interruptible(&dev->outq);
	wake_up_interruptible(&dev->closeq);
    }
    if (int_status & AMCC_BIT_INTCSR_ISRDINT) { /* read is done */
	PDEBUGG("isr: read terminated\n");
	dev->dma_count++;
        dev->rpending=0;
	dev->flags &= ~ADL_FLAG_READING;
	/* tell the device it's done and acknowledge IRQ */
	ADL_OUT32_A(dev, master_status & ~AMCC_BIT_MSCR_DORDREQ,
		    AMCC_REG_MCSR);
	ADL_OUT32_A(dev, int_status | AMCC_BIT_INTCSR_ISMYINT
		    | AMCC_BIT_INTCSR_ISRDINT, AMCC_REG_INTCSR);
	wake_up_interruptible(&dev->inq);
	wake_up_interruptible(&dev->closeq);
    }
    if (int_status & AMCC_BIT_INTCSR_ISMAINT) { /* master abort */
	PDEBUG("isr: master abort\n");
	ADL_OUT32_A(dev, int_status | AMCC_BIT_INTCSR_ISMYINT
		    | AMCC_BIT_INTCSR_ISMAINT, AMCC_REG_INTCSR);
    }
    if (int_status & AMCC_BIT_INTCSR_ISTAINT) { /* target abort */
	PDEBUG("isr: target abort\n");
	ADL_OUT32_A(dev, int_status | AMCC_BIT_INTCSR_ISMYINT
		    | AMCC_BIT_INTCSR_ISTAINT, AMCC_REG_INTCSR);
    }
    int_status = ADL_IN32_L(dev,ADL_REG_DIO_CTRL);
    if (int_status & ADL_BIT_OVER) {
        PDEBUG("isr: overrun\n");
        dev->fifo_overr_count++;
    }
    if (int_status & ADL_BIT_UNDE)  {
        PDEBUG("isr: underrun\n");
        dev->fifo_under_count++;
    }
    int_status |= ADL_BIT_OVER | ADL_BIT_UNDE;
    int_status &= ADL_MASK_DISABLE_ALL;         /* disable: its' done */
    ADL_OUT32_L(dev,int_status,ADL_REG_DIO_CTRL);

    return;
}


/*------------------------------------------------------------ up and down */
int adl_bringup(Adl_Dev *dev)
{
    int result;

    PDEBUG("bringup dev %i\n",dev-adl_devices);
    /* This is comes from ADLink sources: don't know what it is */
    ADL_OUT32(dev, 0x0f000000, AMCC_REG_MCSR);
    ADL_OUT32(dev, 0x00000000, AMCC_REG_MCSR);

    result = request_irq(dev->irq, adl_interrupt, SA_INTERRUPT & SA_SHIRQ,
			 "adl7200", dev);
    if (result) {
	PDEBUG("irq %i not available\n",dev->irq);
	adl_shutdown(dev);
        return result;
    }
    dev->irq_active = dev->irq;
    PDEBUG("got irq number\n");

    if (!(dev->flags & ADL_FLAG_PERSIST)) { /* set defaults */
	dev->flags =              ADL_DEFAULT_FLAGS;
	dev->hertz =              ADL_DEFAULT_HZ;
        dev->counter =            ADL_DEFAULT_COUNTER;
	dev->dma_requested_size = ADL_DEFAULT_DMA_SIZE;
	dev->dma_count = dev->fifo_overr_count = dev->fifo_under_count = 0;
	PDEBUG("defaults loaded, and counters reset\n");
    }

    return 0;

}

int adl_shutdown(Adl_Dev *dev)
{
    PDEBUG("shutdown dev %i\n",dev-adl_devices);
    del_timer(&dev->timer);
    if (dev->softbuffer) {
	free_pages(dev->softbuffer, ADL_SFT_ORDER);
	dev->softbuffer = 0;
    }
    /*
     * Even if DMA can be still active, shutdown the device completely
     * This means resetting the flags, disabling interrupts and releasign
     * the IRQ handler. However, the DMA buffer should not be freed
     * FIXME: on shutdown should abort any DMA, but allow delaying if needed
     */

    /* disable interrupts on the device */
    ADL_OUT32(dev, 0x00000000, AMCC_REG_INTCSR);
    dev->flags &= ~(ADL_FLAG_READING | ADL_FLAG_WRITING);
    if (dev->irq_active) {
	free_irq(dev->irq_active,dev);
    }
    dev->wpending=0;
    dev->rpending=0;
    /* remove timer, just in case */
    del_timer(&dev->timer);

    return 0;
}

/*---------------------------------------------------- the proc filesystem */
#ifdef ADL_USE_PROC

int adl_read_proc(char *buf, char **start, off_t offset,
                   int len, int unused)
{
    int i;
    Adl_Dev *d;

    #define LIMIT (PAGE_SIZE-80) /* don't print anymore after this size */
    len=0;
    for(d=adl_devices; d-adl_devices < ADL_MAX_DEV; d++) {
        u32 dword;
        u16 cmd;

        if (d->io == 0)
	    break;
        
	if (!d->usage)
	    if (adl_bringup(d))
		return len; /* error */
        d->usage++;
	MOD_INC_USE_COUNT;

        len += sprintf(buf+len,
		       "\n\nDevice at %p (bus %i dev %i)\n",
		       d, d->bus, d->device);
#if 0
        len += sprintf(buf+len,"PCI Registers:\n");
        for (i=0; i<0x40; i+=4) {
            if (i%16 == 0)
                len += sprintf(buf+len, "0x%02x: ",i);
            pcibios_read_config_dword(d->bus, d->device, i, &dword);
	    /* cut to 8 bits, as most of them are just 8-bit wide */
            len += sprintf(buf+len, "0x%08x%c", dword,
			   (i+4)%16 ? ' ' : '\n');
            if (len > LIMIT) return len;
        }
#endif
        /* Save the current settings, enable memory I/O and snooze it */

        pcibios_read_config_word(d->bus, d->device, PCI_COMMAND, &cmd);
        pcibios_write_config_word(d->bus, d->device, PCI_COMMAND,
                                  cmd | PCI_COMMAND_IO);

	len += sprintf(buf+len,"\nLocal registers at 0x%04x:\n",(int)d->io);
	for (i=0; i<ADL_MAX_REG; i+=4) {
            if (i%16 == 0)
                len += sprintf(buf+len, "0x%02x: ",i);
            dword=ADL_IN32(d,i);
            len += sprintf(buf+len, "0x%08x%c", dword,
			   (i+4)%16 ? ' ' : '\n');
            if (len > LIMIT) return len;
        }

	/*
	 * it looks like things lock completely by reading all of them,
	 * so read only the last ones. 0x20 locks the computer, so pad it
	 * and start from MWAR
	 */
	len += sprintf(buf+len,"\nAMCC registers at 0x%04x:\n",
                       (int)d->io_amcc);
        len += sprintf(buf+len,"0x20:   -------- ");

        for (i=AMCC_REG_MWAR; i<AMCC_MAX_REG; i+=4) {
            if (i%16 == 0)
                len += sprintf(buf+len, "0x%02x: ",i-AMCC_MIN_REG);
            dword=ADL_IN32(d,i);
            len += sprintf(buf+len, "0x%08x%c", dword,
			   (i+4)%16 ? ' ' : '\n');
            if (len > LIMIT) return len;
        }
	/* done */

        pcibios_write_config_word(d->bus, d->device, PCI_COMMAND, cmd);
        d->usage--;
	if (!d->usage)
	    adl_shutdown(d);
	MOD_DEC_USE_COUNT;
    }
    #undef LIMIT
    return len;
}

struct proc_dir_entry adl_proc_entry = {
        0,                 /* low_ino: the inode -- dynamic */
        7, "adl7200",      /* len of name and name */
        S_IFREG | S_IRUGO, /* mode */
        1, 0, 0,           /* nlinks, owner, group */
        0, NULL,           /* size - unused; operations -- use default */
        &adl_read_proc,     /* function used to readl data */
        /* nothing more */
    };
#endif

/*---------------------------------------------------------- allocate DMA */
void adl_free_dma(Adl_Dev *dev)
{
    if (dev->dmaremap) {
	iounmap((void *)dev->dmaremap);
	dev->dmaremap=0;
    }
    if (dev->dmabuffer) {
        PDEBUG("freeing DMA buffer\n");
	allocator_free_dma(dev->dmabuffer);
	dev->dmabuffer =0;
        PDEBUG("freed DMA buffer\n");
    }
    dev->dmasize = 0;
    /* don't override other values, which are useful in reallocating */
}

int adl_allocate_dma(Adl_Dev *dev, int kilobytes, int prio)
{
    if (dev->dmasize >= kilobytes * 1024)
        return 0;
    if (dev->dmasize)
        adl_free_dma(dev);
    dev->dmabuffer = allocator_allocate_dma(kilobytes, prio);
    PDEBUG("allocating: %i kilos, got 0x%08lx\n", kilobytes,dev->dmabuffer);
    if (!dev->dmabuffer)
	return -ENOMEM;
    dev->dmasize = kilobytes << 10;
    dev->dmaremap = (unsigned long) ioremap(dev->dmabuffer, dev->dmasize);
    return 0;
}

/*---------------------------------------------------------- set the clock */
/* This function doesn't play with the AMCC, only with add-on logic */
int adl_set_clock(Adl_Dev *dev)
{
    u32 tmp;

    if (dev->counter > 0xffff) {
	PDEBUG("counter out of range\n");
	return -EINVAL;
    }
    tmp =ADL_IN32_L(dev, ADL_REG_IRQ_CTRL);
    ADL_OUT32_L(dev,tmp & ~(ADL_BIT_CASCADE01 | ADL_BIT_CASCADE02),
		ADL_REG_IRQ_CTRL);
    
    ADL_OUT32_L(dev, 0x74 /* mode 2 */, ADL_REG_CNT_CTRL);
    ADL_OUT32_L(dev, dev->counter & 0xff, ADL_REG_CNT1);
    ADL_OUT32_L(dev, dev->counter >> 8,   ADL_REG_CNT1);

    ADL_OUT32_L(dev, 0x34 /* mode 2 */, ADL_REG_CNT_CTRL);
    ADL_OUT32_L(dev, dev->counter & 0xff, ADL_REG_CNT0);
    ADL_OUT32_L(dev, dev->counter >> 8,   ADL_REG_CNT0);

    return 0;
}

/*---------------------------------------------------------- open */

int adl_open (struct inode *inode, struct file *filp)
{
    int type = TYPE(inode->i_rdev);
    int num = NUM(inode->i_rdev);
    int retval = 0;
    Adl_Dev *dev;

    if (type > ADL_MAX_TYPE) return -ENODEV;
    if (num >= ADL_MAX_DEV)  return -ENODEV;

    dev = adl_devices + num;
    if (!dev->io)
	return -ENODEV;

    filp->private_data = dev;
    filp->f_op = adl_fops[type];

    /* Ok, now bring it up */
    if (!dev->usage)
	if ( (retval=adl_bringup(dev)) )
	    return retval; /* error */
    else if (dev->flags & ADL_FLAG_SINGLEO)
	return -EBUSY;

    dev->usage++;
    MOD_INC_USE_COUNT;

    if (type != 0) /* else, fall through */
	if (filp->f_op->open)
	    retval = (filp->f_op->open)(inode,filp); /* special open code */
	else {
	    retval = -ENODEV; /* no open method: not implemented */
	}
    else {
	/* type is 0: nothing to do */
	retval = 0;
    }
    if (retval) {
	dev->usage--;
	if (!dev->usage)
	    adl_shutdown(dev);
	MOD_DEC_USE_COUNT;
    }
    return retval;
}

/*-------------------------------------------------------- close */

release_t adl_close (struct inode *inode, struct file *filp)
{
    Adl_Dev *dev = filp->private_data;

    PDEBUG("close: usage %i (mod %li)\n",dev->usage,
	   (mod_use_count_ & 0xfff) );
    /* FIXME: implement blocking close */
    MOD_DEC_USE_COUNT;
    dev->usage--;
    if (dev->usage==0) {
	if (dev->irq_active) {
	    free_irq(dev->irq_active, dev);
	    dev->irq_active = 0;
	}
	adl_shutdown(dev);
	if (!(dev->flags & (ADL_FLAG_READING | ADL_FLAG_WRITING)))
	  adl_free_dma(dev); /* FIXME: use a kernel timer instead */
    }
    release_return(0);
}

/*-------------------------------------------------------- read */

read_write_t adl_read (struct inode *inode, struct file *filp,
                char *buf, count_t count)
{
    Adl_Dev *dev = filp->private_data;
    count_t xferred=0;
    u32 datum=0; /* Bah! */

    /*
     * Return a continuous flow of 32bit values.
     */
    while(xferred < count-3) {
	datum=ADL_IN32(dev,ADL_REG_DIN);
	put_user(datum,(u32 *)buf);
	xferred += sizeof(u32);
	buf += sizeof(u32);
    }
    filp->f_pos += xferred;
    return xferred; /* no, can't read one byte at a time */
}

/*-------------------------------------------------------- write */

read_write_t adl_write (struct inode *inode, struct file *filp,
                const char *buf, count_t count)
{
    Adl_Dev *dev = filp->private_data;
    count_t xferred=0;
    u32 datum;

    /*
     * Send out a continuous flow of 32bit values.
     */
    while(xferred < count-3) {
	GET_USER(datum,(u32 *)buf);
	ADL_OUT32(dev, datum, ADL_REG_DOUT);
	xferred += sizeof(u32);
	buf += sizeof(u32);
    }
    filp->f_pos += xferred;
    if (count < sizeof(u32))
	return count; /* fake we got them, so redirection works */
    return xferred;
}

/*-------------------------------------------------------- ioctl */
int adl_ioctl (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
    Adl_Dev *dev = filp->private_data;
    int size = _IOC_SIZE(cmd); /* the size bitfield in cmd */
    int retval = 0; /* success by default */
    #define IOC_BUFSIZE 32 /* some space, just random */
    static unsigned char localbuf[IOC_BUFSIZE];
    /* I'm lazy, and I define "alias" names for the buffer */
    void *karg = localbuf;
    unsigned long *klong = karg;
    struct Adl_Io_Struct *kios = karg;

    /*
     * extract the type and number bitfields, and don't decode
     * wrong cmds: return EINVAL before verify_area()
     */
    if (_IOC_TYPE(cmd) != ADL_IOC_MAGIC) return -EINVAL;
    if (_IOC_NR(cmd) > ADL_IOC_MAXNR) return -EINVAL;

    /*
     * the direction is a bitmask, and VERIFY_WRITE catches R/W
     * transfers. `Dir' is user-oriented, while
     * verify_area is kernel-oriented, so the concept of "read" and
     * "write" is reversed
     */
    if (_IOC_DIR(cmd) & _IOC_READ) {
        if (!access_ok(VERIFY_WRITE, (void *)arg, size))
            return -EFAULT;
    }
    else if (_IOC_DIR(cmd) & _IOC_WRITE) {
        if (!access_ok(VERIFY_READ, (void *)arg, size))
            return -EFAULT;
    }

    PDEBUGG("ioctl: %08x %08lx: nr %i, size %i, dir %i\n",
            cmd, arg, _IOC_NR(cmd), _IOC_SIZE(cmd), _IOC_DIR(cmd));

    /* first, retrieve data from userspace */
    if ((_IOC_DIR(cmd) & _IOC_WRITE) && (size <= IOC_BUFSIZE))
        copy_from_user(karg, (void *)arg, size);

    /* and now the big switch... */
    switch(cmd) {
     case ADL_IOCHARDRESET: /* bring use count to 1, to allow unloading */
        while (MOD_IN_USE)
	    MOD_DEC_USE_COUNT;
	MOD_INC_USE_COUNT;
        dev->usage = 1;
        /* fall through */
      case ADL_IOCRESET:                    /* reset the device */
        adl_shutdown(dev);
	dev->dma_count = dev->fifo_overr_count = dev->fifo_under_count = 0;
        retval = adl_bringup(dev);
        break;

	/* FIXME: ioctl for in and out */

        /*
         * reading and writing values:
         * all of them are "unsigned long"
         */

      case ADL_IOCGFLAGS:                   /* get flags */
        *klong = dev->flags;
        break;

      case ADL_IOCSFLAGS:                   /* set flags */
	if (dev->flags & (ADL_FLAG_READING | ADL_FLAG_WRITING))
            return -EBUSY;
        dev->flags = *klong & ~(ADL_FLAG_READING | ADL_FLAG_WRITING);
        break;

      case ADL_IOCGTIMER:                   /* get timer */
        *klong = dev->counter;
        break;

      case ADL_IOCSTIMER:                   /* set timer */
	if (*klong >= 1<<16) {
            PDEBUG("ioctl: can't set timer to %li\n",*klong);
            return -EINVAL;
        }
        dev->counter = *klong;
        break;

      case ADL_IOCGHERTZ:                   /* get hertz */
        *klong = dev->hertz;
        break;

      case ADL_IOCSHERTZ:                   /* set hertz */
        if (*klong > HZ) {
            PDEBUG("ioctl: can't set hertz to %li\n",*klong);
            return -EINVAL;
        }
        dev->hertz = *klong;
        break;

      case ADL_IOCGDMASIZE:                 /* get dmasz */
        *klong = dev->dma_requested_size;
        break;

      case ADL_IOCSDMASIZE:                 /* set dmasz */
        if (dev->flags & (ADL_FLAG_READING | ADL_FLAG_WRITING))
            return -EBUSY;
	PDEBUG("calling free\n");
	adl_free_dma(dev);
	PDEBUG("free called\n");
        dev->dma_requested_size = *klong;
        break;

      case ADL_IOCGDMACOUNT:                /* get dma count */
        *klong = dev->dma_count;
        break;

      case ADL_IOCGUNDERRUN:                /* get nr. underrun */
        *klong = dev->fifo_under_count;
        break;

      case ADL_IOCGOVERRUN:                 /* get nr. overrun */
        *klong = dev->fifo_overr_count;
        break;
    }

    /* finally, copy data to user space and return */
    if (retval < 0) return retval;
    if ((_IOC_DIR(cmd) & _IOC_READ) && (size <= IOC_BUFSIZE))
        copy_to_user((void *)arg, karg, size);
    return retval; /* sometimes, positive is what I want */
}


/*-------------------------------------------------------- select */
#ifdef __USE_OLD_SELECT__
int adl_poll (struct inode *inode, struct file *filp,
                  int mode, select_table *table)
{
    Adl_Dev *dev = filp->private_data;

    switch(NUM(inode->i_rdev)) {
      case ADL_RAW:
        return 1; /* readable, writable and exceptionable */
      default:
        return 1; /* not implemented: default to true */
    }
}
#else
unsigned int adl_poll (struct file *filp, poll_table *table)
{
    int ret=0;

    Adl_Dev = filp->private_data;

    poll_wait(&dev->inq,  table);
    poll_wait(&dev->outq, table);
    poll_wait(&dev->closeq, table);
    switch(NUM(inode->i_rdev)) {
      case ADL_RAW: /* readable, writable, exceptionable */
        return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM | POLLPRI;
      default: /* not implemented, default to anything */
        return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM | POLLPRI;
    }
    return ret;
}
#endif

/*-------------------------------------------------------- the fops */

/* The raw node */
struct file_operations adl_raw_fops = {
    NULL,          /* adl_lseek */
    adl_read,
    adl_write,
    NULL,          /* adl_readdir */
    adl_poll,
    adl_ioctl,
    NULL,          /* adl_mmap */
    adl_open,
    adl_close,
    NULL,          /* adl_fsync */
    NULL,          /* adl_fasync */
                   /* nothing more, fill with NULLs */
};

/* dma and trig ops are in dma.c */

/* soft ops are in soft.c */

struct file_operations adl_asc_fops = {
    NULL,          /* adl_lseek */
    adl_read,
    adl_write,
    NULL,          /* adl_readdir */
    NULL,          /* adl_select */
    NULL,          /* adl_ioctl */
    NULL,          /* adl_mmap */
    NULL,     /* open */
    adl_close,
    NULL,          /* adl_fsync */
    NULL,          /* adl_fasync */
                   /* nothing more, fill with NULLs */
};


/*
 * Finally, the module stuff.
 */

/*-------------------------------------------------------- modules */
int init_module(void)
{
    int result, i;
    struct Adl_Dev *ptr;

    /*
     * This will leadl to an insmod error if CONFIG_PCI was not set in the
     * target machine.
     */
    if (!pcibios_present()) {
        printk(KERN_ERR ADL_MSG "PCI bios needed: can't install\n");
        return -ENODEV;
    }

    /*
     * First of all allocate the device structures
     */

    for (i=0, ptr = adl_devices; i < ADL_MAX_DEV; ptr++, i++) {
        u8 bus, device;

	/* look for the board */
        result = pcibios_find_device(ADL_VENDOR_ID, ADL_DEVICE_ID,
				     i, &bus, &device);
	if (result) {
	    if (result == PCIBIOS_DEVICE_NOT_FOUND) {
	        result = 0; break; /* this is good */
	    } 
	    printk(KERN_ERR ADL_MSG "pcibios: %s\n",
		     pcibios_strerror(result));
	    result = -EIO;
	    break;
	}

	/* so, one is there */	
        PDEBUG("found at bus %i device %i\n",bus,device);

        memset(ptr,0,sizeof(Adl_Dev));
	ptr->bus = bus;
	ptr->device = device;

	if ( (result = adl_init(ptr)) ) {
	    if (result == -ENODEV) {
	        PDEBUG("... not a 7200\n");
	        continue; /* try again */
	    }
	    else {
	        PDEBUG("initialization failed\n");
		break;
	    }
	}
    }

    if (i==0) {
        printk(KERN_ERR ADL_MSG "no device found\n");
        return -ENODEV;
    }
       
    /*
     * Register your major, and accept a dynamic number
     */
    result = register_chrdev(adl_major, "adl7200", adl_fops[0]);
    if (result < 0) {
        printk(KERN_WARNING ADL_MSG "can't get major %d\n",adl_major);
	return result;
    }
    if (adl_major == 0) adl_major = result; /* dynamic */

#ifndef ADL_DEBUG
    register_symtab(NULL); /* otherwise, leave global symbols visible */
#endif

#ifdef ADL_USE_PROC
    proc_register_dynamic(&proc_root, &adl_proc_entry);
#endif

    allocator_init();
    printk(KERN_INFO ADL_MSG "driver " ADL_VERSION_STRING " loaded\n");
    PDEBUG("Compiled on " __DATE__ ", " __TIME__ "\n");
    return 0; /* succeed */
}

void cleanup_module(void)
{
    unregister_chrdev(adl_major, "adl7200");
    allocator_cleanup();
#ifdef ADL_USE_PROC
    proc_unregister(&proc_root, adl_proc_entry.low_ino);
#endif
    printk(KERN_INFO "adlc200: driver unloadled\n");
}
